#!/usr/bin/env python

"""
Script to handle the production of DL3 files.

It uses the lstchain Tools:
 - lstchain_create_irf_files
 - lstchain_create_dl3_file
 - lstchain_create_dl3_index_files
"""

import logging
import subprocess as sp
from datetime import datetime
from pathlib import Path

import click
from astropy.utils import iers

from osa.configs import options
from osa.configs.config import cfg
from osa.nightsummary.extract import build_sequences, get_source_list
from osa.paths import destination_dir, DEFAULT_CFG, create_source_directories, analysis_path
from osa.utils.cliopts import get_prod_id, get_dl2_prod_id
from osa.utils.logging import myLogger
from osa.utils.utils import stringify, YESTERDAY

iers.conf.auto_download = False

__all__ = [
    "batch_cmd_create_irf",
    "batch_cmd_create_dl3",
    "batch_cmd_create_index_dl3",
    "get_irf_file",
    "create_irf",
    "produce_dl3_files",
    "setup_global_options",
    "cuts_subdirectory",
    "create_obs_index",
]

log = myLogger(logging.getLogger(__name__))


def batch_cmd_create_irf(cwd, mc_gamma, mc_proton, mc_electron, output_irf_file, dl3_config):
    """Create batch command to create IRF file with sbatch."""
    return [
        "sbatch",
        "--parsable",
        "--mem=6GB",
        "--job-name=irf",
        "-D",
        cwd,
        "-o",
        "log/create_irf_%j.log",
        "lstchain_create_irf_files",
        "--point-like",
        f"--input-gamma-dl2={mc_gamma}",
        f"--input-proton-dl2={mc_proton}",
        f"--input-electron-dl2={mc_electron}",
        f"--output-irf-file={output_irf_file}",
        f"--config={dl3_config}",
        "--overwrite",
    ]


def batch_cmd_create_dl3(
    dl2_file, dl3_dir, run, source_name, source_ra, source_dec, irf, dl3_config, job_irf
):
    """Create batch command to create DL3 files with sbatch."""
    log_dir = dl3_dir / "log"
    log_dir.mkdir(exist_ok=True, parents=True)
    log_file = log_dir / f"dl2_to_dl3_Run{run:05d}_{source_name}_%j.log"
    sbatch_cmd = [
        "sbatch",
        "--mem=10GB",
        "--job-name=dl2dl3",
        "-D",
        dl3_dir,
        "-o",
        log_file,
        "--parsable",
    ]
    if job_irf is not None:
        sbatch_cmd.append(f"--dependency=afterok:{job_irf}")

    dl3_cmd = [
        "lstchain_create_dl3_file",
        f"-d={dl2_file}",
        f"-o={dl3_dir}",
        f"--input-irf={irf}",
        f"--source-name={source_name}",
        f"--source-ra={source_ra}deg",
        f"--source-dec={source_dec}deg",
        "--overwrite",
    ]
    if dl3_config:
        dl3_cmd.append(f"--config={dl3_config}")

    return sbatch_cmd + dl3_cmd


def batch_cmd_create_index_dl3(dl3_dir, parent_job_list):
    """Build batch command to create the observations index."""
    parent_job_list_str = ",".join(parent_job_list)
    log_dir = dl3_dir / "log"
    log_dir.mkdir(exist_ok=True, parents=True)
    log_file = log_dir / "create_index_dl3_%j.log"

    return [
        "sbatch",
        "--parsable",
        "--mem=8GB",
        "--job-name=dl3_index",
        f"--dependency=afterok:{parent_job_list_str}",
        "-D",
        dl3_dir,
        "-o",
        log_file,
        "lstchain_create_dl3_index_files",
        f"-d={dl3_dir}",
        f"-o={dl3_dir}",
        "-p=dl3*.fits.gz",
        "--overwrite",
    ]


def get_irf_file(directory: Path, simulate: bool = False):
    """Return the irf file. Create it if not existing."""
    if cfg.get("MC", "IRF_file") is not None:
        irf_file = Path(cfg.get("MC", "IRF_file"))
        log.info(f"Using existing IRF file:\n{irf_file}")
        dl3_config = None  # Uses default settings
        job_id_irf = None
    else:
        dl3_config = Path(cfg.get("lstchain", "DL3_CONFIG"))
        irf_file, job_id_irf = create_irf(directory, dl3_config, simulate)

    return irf_file, dl3_config, job_id_irf


def create_irf(directory: Path, config: Path, simulate: bool = False):
    """Create the IRF file for a given set of selection cuts."""
    log.info("Creating the IRFs.")

    mc_gamma = cfg.get("MC", "gamma")
    mc_proton = cfg.get("MC", "proton")
    mc_electron = cfg.get("MC", "electron")
    irf_file = directory / "irf.fits.gz"

    cmd1 = batch_cmd_create_irf(
        cwd=directory,
        mc_gamma=mc_gamma,
        mc_proton=mc_proton,
        mc_electron=mc_electron,
        output_irf_file=irf_file,
        dl3_config=config,
    )

    if simulate:
        return None

    log.info("Submitting the IRF job.")
    log.debug(stringify(cmd1))
    job_irf = sp.run(
        cmd1,
        encoding="utf-8",
        capture_output=True,
        text=True,
    )
    return irf_file, job_irf.stdout.strip()


def produce_dl3_files(
    sequence_list,
    irf_file: Path,
    dl2_dir: Path,
    cuts_dir: Path,
    dl3_config: Path,
    job_id_irf: int,
    simulate: bool = False,
):
    """Produce the DL3 files for a given list of sequences."""
    list_of_job_id = []

    log.info("Looping over the runs to produce the DL3 files.")
    for sequence in sequence_list:

        if sequence.type == "DATA":

            dl2_file = dl2_dir / f"dl2_LST-1.Run{sequence.run:05d}.h5"
            dl3_subdir = cuts_dir / f"{sequence.source_name}"

            cmd2 = batch_cmd_create_dl3(
                dl2_file=dl2_file,
                dl3_dir=dl3_subdir,
                run=sequence.run,
                source_name=sequence.source_name,
                source_ra=sequence.source_ra,
                source_dec=sequence.source_dec,
                irf=irf_file,
                dl3_config=dl3_config,
                job_irf=job_id_irf,
            )

            if not simulate:
                log.info(f"Producing DL3 file for run {sequence.run:05d}")
                job_id = sp.run(
                    cmd2,
                    encoding="utf-8",
                    capture_output=True,
                    text=True,
                )

                list_of_job_id.append(job_id.stdout.strip())

            else:
                log.debug("Simulate launching scripts")

            log.debug(f"Executing {stringify(cmd2)}")

    return list_of_job_id


def create_obs_index(source_list: list, cuts_dir: Path, parent_jobs: list, simulate: bool = False):
    """Creating observation index for each source."""
    log.info("Creating observation index for each source.")

    for source in source_list:
        dl3_subdir = cuts_dir / source

        cmd3 = batch_cmd_create_index_dl3(dl3_subdir, parent_jobs)

        if not simulate:
            log.info("Scheduling DL3 index job")
            sp.run(
                cmd3,
                encoding="utf-8",
                capture_output=True,
                text=True,
            )
        else:
            log.debug("Simulate creating DL3 index")

        log.debug(f"Executing {stringify(cmd3)}")


def setup_global_options(date_obs, telescope):
    """Set up the global options arguments."""
    options.date = date_obs
    options.tel_id = telescope
    options.prod_id = get_prod_id()
    options.dl2_prod_id = get_dl2_prod_id()
    options.directory = analysis_path(options.tel_id)


def cuts_subdirectory() -> Path:
    """Return the dl3 subdirectory for a given set of cuts."""
    log.debug("Creating DL3 directory")
    dl3_dir = destination_dir("DL3", create_dir=True)
    std_cuts_dir = dl3_dir / "std_cuts"
    log_dir = std_cuts_dir / "log"
    std_cuts_dir.mkdir(parents=True, exist_ok=True)
    log_dir.mkdir(exist_ok=True)
    log.debug(f"DL3 directory: {dl3_dir}")
    return std_cuts_dir


@click.command()
@click.argument("telescope", type=click.Choice(["LST1", "LST2"]))
@click.option("-d", "--date-obs", type=click.DateTime(formats=["%Y-%m-%d"]), default=YESTERDAY)
@click.option(
    "-c",
    "--config",
    type=click.Path(dir_okay=False),
    default=DEFAULT_CFG,
    help="Read option defaults from the specified cfg file",
)
@click.option("-v", "--verbose", is_flag=True)
@click.option("--local", is_flag=True)
@click.option("-s", "--simulate", is_flag=True)
def main(
    date_obs: datetime = YESTERDAY,
    telescope: str = "LST1",
    verbose: bool = False,
    simulate: bool = False,
    local: bool = False,
    config: Path = DEFAULT_CFG,
):
    """Produce the IRF and DL3 files tool in a run basis."""
    log.setLevel(logging.INFO)

    if verbose:
        log.setLevel(logging.DEBUG)

    log.info(f"=== DL3 stage for {date_obs.strftime('%Y-%m-%d')} ===")
    log.debug(f"Config: {config.resolve()}")

    if local:
        options.test = True
        log.info("Local mode enabled: no interaction with the cluster.")

    if simulate:
        options.simulate = True
        log.info("Simulation mode enabled: no jobs will be submitted.")

    # Set up the global options
    setup_global_options(date_obs, telescope)
    dl2_dir = destination_dir("DL2", create_dir=False)

    # Build the sequences
    sequence_list = build_sequences(options.date)

    # Get the list of source names
    source_list = list(get_source_list(options.date))

    # Create a subdirectory inside the DL3 directory corresponding to the selection cuts
    std_cuts_dir = cuts_subdirectory()

    # Create a subdirectory for each source
    create_source_directories(source_list, std_cuts_dir)

    # Get IRF file
    irf_file, dl3_config, job_id_irf = get_irf_file(directory=std_cuts_dir, simulate=simulate)

    # Create the DL3 files
    list_of_job_id = produce_dl3_files(
        sequence_list=sequence_list,
        irf_file=irf_file,
        dl2_dir=dl2_dir,
        cuts_dir=std_cuts_dir,
        dl3_config=dl3_config,
        job_id_irf=job_id_irf,
        simulate=simulate,
    )

    # Creating an observation file index for each source
    create_obs_index(
        source_list=source_list,
        cuts_dir=std_cuts_dir,
        parent_jobs=list_of_job_id,
        simulate=simulate,
    )


if __name__ == "__main__":
    main()



"""Produce the HTML file with the processing status from the sequencer report."""


import logging
import subprocess as sp
import sys
from datetime import datetime, timedelta
from pathlib import Path
from textwrap import dedent
from typing import Iterable

import pandas as pd

from osa.configs import options
from osa.configs.config import cfg
from osa.utils.cliopts import sequencer_webmaker_argparser
from osa.utils.logging import myLogger
from osa.utils.utils import is_day_closed, date_to_iso, date_to_dir

log = myLogger(logging.getLogger())


def html_content(body: str, date: str) -> str:
    """Build the HTML content.

    Parameters
    ----------
    body : str
        Table with the sequencer status report.
    date : str
        Date of the processing YYYY-MM-DD.

    Returns
    -------
    str
        HTML content.
    """
    time_update = datetime.utcnow().strftime("%Y-%m-%d %H:%M:%S")

    return dedent(
        f"""<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Strict//EN"
        "http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd">
        <html xmlns="http://www.w3.org/1999/xhtml">
         <head>
          <meta http-equiv="Content-Type" content="text/html; charset=utf-8" />
          <title>OSA Sequencer status</title><link href="osa.css" rel="stylesheet"
          type="text/css" /><style>table{{width:152ex;}}</style>
         </head>
         <body>
         <h1>OSA processing status</h1>
         <p>Processing data from: {date}. Last updated: {time_update} UTC</p>
         {body}
         </body>
        </html>"""
    )


def get_sequencer_output(date: str, config: str, test=False) -> list:
    """Call sequencer to get table with the sequencer status report.

    Parameters
    ----------
    date : str
        Date of the processing YYYY-MM-DD.
    config : str
        OSA configuration file to use.
    test : bool

    Returns
    -------
    list
        Lines of the sequencer output.
    """
    log.info("Calling sequencer...")
    commandargs = [
        "sequencer",
        "-c",
        config,
        "-s",
        "-d",
        date,
        options.tel_id,
    ]

    if test:
        commandargs.insert(-1, "-t")

    try:
        output = sp.run(commandargs, stdout=sp.PIPE, stderr=sp.STDOUT, encoding="utf-8", check=True)
    except sp.CalledProcessError as error:
        log.error(f"Command {commandargs} failed, {error.returncode}")
        sys.exit(1)
    else:
        # Strip newlines and fit it into a table:
        return output.stdout.splitlines()


def lines_to_matrix(lines: Iterable) -> list:
    """Build the matrix from the sequencer output lines."""
    matrix = []
    for line in lines:
        l_fields = line.split()
        if len(l_fields) == 18:
            matrix.append(l_fields)
    return matrix


def matrix_to_html(matrix: list) -> str:
    """Build the html table with the sequencer status report."""
    log.info("Building the html table from sequencer output")
    if len(matrix) < 2:
        return "<p>No data found</p>"
    df = pd.DataFrame(matrix[1:], columns=matrix[0])
    return df.to_html(index=False)


def main():
    """Produce the html file with the processing status from the sequencer report."""

    log.setLevel(logging.INFO)

    args = sequencer_webmaker_argparser().parse_args()

    if args.date:
        flat_date = date_to_dir(args.date)
        options.date = args.date

    else:
        # yesterday by default
        yesterday = datetime.now() - timedelta(days=1)
        options.date = yesterday
        flat_date = date_to_dir(yesterday)

    date = date_to_iso(options.date)

    if is_day_closed():
        log.info(f"Date {date} is already closed for {options.tel_id}")
        sys.exit(1)

    run_summary_directory = Path(cfg.get("LST1", "RUN_SUMMARY_DIR"))
    run_summary_file = run_summary_directory / f"RunSummary_{flat_date}.ecsv"
    if not run_summary_file.is_file():
        log.error(f"No RunSummary file found for {date}")
        sys.exit(1)

    # Get the table with the sequencer status report:
    lines = get_sequencer_output(date, args.config, args.test)

    # Build the html sequencer table that will be place in the body of the HTML file
    matrix = lines_to_matrix(lines)
    html_table = matrix_to_html(matrix)

    # Save the HTML file
    log.info("Saving the HTML file")
    directory = Path(cfg.get("LST1", "SEQUENCER_WEB_DIR"))
    directory.mkdir(parents=True, exist_ok=True)

    html_file = directory / Path(f"osa_status_{flat_date}.html")
    html_file.write_text(html_content(html_table, date), encoding="utf-8")

    log.info("Done")


if __name__ == "__main__":
    main()



import logging

from osa.configs import options
from osa.utils.utils import gettag, stringify

log = logging.getLogger(__name__)


def select_db(servername, username, database, table, selections, conditions):
    appendix = None
    feedback = select_appendix_db(
        servername, username, database, table, selections, conditions, appendix
    )
    return feedback


def select_appendix_db(servername, username, database, table, selections, conditions, appendix):
    tag = gettag()

    verb = select_selections(selections)
    assignments = {}
    feedback = query_db(
        servername, username, database, table, verb, assignments, conditions, appendix
    )
    """ The returned object will be a matrix: list of lists """
    matrix = []
    if feedback:
        for line in feedback.splitlines():
            fields = line.split("\t")
            if len(fields) == len(selections):
                # Allright
                pass
            else:
                log.error("Wrong number of selected items {0}".format(fields))
            matrix.append(fields)
    return matrix


def insert_and_select_id_db(servername, username, database, table, assignments, conditions):
    tag = gettag()

    id = None
    verb = "INSERT IGNORE"
    appendix = "; SELECT LAST_INSERT_ID()"
    feedback = query_db(
        servername, username, database, table, verb, assignments, conditions, appendix
    )
    """ If simulation, better have a method to return 0 anyway """
    if not feedback:
        id = 0
    else:
        id = int(feedback.strip())
    log.debug("LAST_INSERT_ID is {0}".format(id))
    return id


def update_and_select_id_db(servername, username, database, table, assignments, conditions):
    tag = gettag()

    id = None
    selections = ["ID"]
    feedback = select_db(servername, username, database, table, selections, conditions)
    if len(feedback) > 0:
        id = int(feedback[0][0])
        conditions = {"ID": id}
        verb = "UPDATE IGNORE"
        appendix = None
        feedback = query_db(
            servername, username, database, table, verb, assignments, conditions, appendix
        )
    else:
        """ If simulation, better have a method to return 0 anyway """
        id = 0
    return id


#############################################################################
#
# update_or_insert_and_select_id_db
#
##############################################################################
def update_or_insert_and_select_id_db(
    servername, username, database, table, assignments, conditions
):
    """ It tries an update of the db and, if it fails, then inserts """
    feedback = None
    id = update_and_select_id_db(servername, username, database, table, assignments, conditions)
    if id == 0:
        """ There was no update, proceed with insertion """
        assignments.update(conditions)
        conditions = {}
        id = insert_and_select_id_db(servername, username, database, table, assignments, conditions)
        log.debug("Inserted with ID={0}".format(id))
    else:
        log.debug("Updated with ID={0}".format(id))
    return id


def insert_db(servername, username, database, table, assignments, conditions):
    tag = gettag()

    verb = "INSERT"
    appendix = None
    feedback = query_db(
        servername, username, database, table, verb, assignments, conditions, appendix
    )
    return feedback


def update_db(servername, username, database, table, assignments, conditions):
    tag = gettag()

    verb = "UPDATE"
    appendix = None
    feedback = query_db(
        servername, username, database, table, verb, assignments, conditions, appendix
    )
    return feedback


def insert_ignore_db(servername, username, database, table, assignments, conditions):
    tag = gettag()

    verb = "INSERT IGNORE"
    appendix = None
    feedback = query_db(
        servername, username, database, table, verb, assignments, conditions, appendix
    )
    return feedback


def update_ignore_db(servername, username, database, table, assignments, conditions):
    tag = gettag()

    verb = "UPDATE IGNORE"
    appendix = None
    feedback = query_db(
        servername, username, database, table, verb, assignments, conditions, appendix
    )
    return feedback


def query_db(servername, username, database, table, verb, assignments, conditions, appendix):
    tag = gettag()
    set_a = set_assignments(assignments)
    where_c = where_conditions(conditions)
    sql_command = "{0} {1} {2} {3}".format(verb, table, set_a, where_c)
    if appendix:
        sql_command += " {0}".format(appendix)
    feedback = connect_or_call(servername, username, database, sql_command)
    return feedback


def select_selections(selections):
    tag = gettag()

    string = phrase_builder("SELECT", ",", selections)
    string += " FROM"
    return string


def set_assignments(assignments):
    tag = gettag()

    string = phrase_builder("SET", ",", assignments)
    return string


def where_conditions(conditions):
    tag = gettag()

    string = phrase_builder("WHERE", " AND", conditions)
    return string


def phrase_builder(instruction, junction, assignments):
    """
    With the assignments as a dictionary, we can compose a sentence,
    starting by instruction and separated by the junction.
    """

    import types

    string = ""
    if len(assignments) != 0:
        phrase = instruction
        # if type(assignments) == types.DictType:
        if type(assignments) == types.MemberDescriptorType:
            for key in assignments:
                if assignments[key] != None:
                    if key == "MD5SUM":
                        # Force checksums to be strings
                        phrase += " {0}='{1}'{2}".format(key, assignments[key], junction)
                    elif is_number(assignments[key]) or is_parenthesis(assignments[key]):
                        phrase += " {0}={1}{2}".format(key, assignments[key], junction)
                    else:
                        phrase += " {0}='{1}'{2}".format(key, assignments[key], junction)
                else:
                    phrase += " {0}=NULL{1}".format(key, junction)
        # elif type(assignments) == types.ListType:
        elif type(assignments) == types.FrameType:
            for element in assignments:
                phrase += " {0}{1}".format(element, junction)
        string = phrase.rstrip(junction)
    return string


def is_number(s):
    tag = gettag()

    """ It allows to check if the entry s is a number or not. """

    try:
        # chapuza to make it slightly more robust:
        # even if we can convert to float
        # it is not a number if the number is absurdly
        # long and there's no decimal point
        float(s)
        assert len(s) < 20 or "." in s
        return True
    except ValueError:
        return False
    except TypeError:
        return False
    except AssertionError:
        return False


def is_parenthesis(s):
    tag = gettag()

    """ It checks wheter the string is enclosed by parenthesis """
    if isinstance(s, str) and len(s) > 0:
        if s[0] == "(" and s[-1] == ")":
            return True
    return False


def connect_or_call(servername, username, database, sql_command):
    tag = gettag()

    """ This is a frontend def to allow making a subprocess call if the api is
    not available, instead of crying because of broken python installation. """

    feedback = None
    if not options.simulate:
        try:
            import MySQLdb
        except ImportError:
            feedback = subprocess_call_db(servername, username, database, sql_command)
        else:
            feedback = connect_db(servername, username, database, sql_command)
    else:
        log.debug("SIMULATE query to database: {0}".format(sql_command))
    return feedback


def subprocess_call_db(servername, username, database, sql_command):
    tag = gettag()
    # Let's do it through a system call
    import subprocess

    feedback = None
    commandargs = ["mysql"]
    if servername:
        commandargs.append("-h")
        commandargs.append(servername)
    if username:
        commandargs.append("-u")
        commandargs.append(username)
    if database:
        commandargs.append("-D")
        commandargs.append(database)
    commandargs.append("--batch")
    commandargs.append("-N")
    commandargs.append("-e")
    """
    if len(sql_command.split(" MD5SUM=")) != 1:
        md5sum_text = str(sql_command.split(' MD5SUM=')[-1].split(',')[0])
        outer  = sql_command.split(md5sum_text)
        #print(md5sum_text)
        #new = "'%s'"% md5sum
        #print(md5sum, new)
        #sql_command.replace(old, inew)
        if "NULL" not in md5sum_text:
            sql_command = str("%s'%s'%s" %(outer[0],md5sum_text,outer[1]))
    """
    commandargs.append(sql_command)

    try:
        feedback = subprocess.check_output(commandargs)
    # except OSError as (ErrorValue, ErrorName):
    except OSError as error:
        log.error(error)
    except subprocess.CalledProcessError as error:
        log.error("MySQL> {0}: {1}".format(stringify(error.cmd), error.output), error.returncode)
    else:
        log.debug("Database query OK: {0}".format(sql_command))
    return feedback


def connect_db(servername, username, database, sql_command):
    tag = gettag()
    # Let's do it through the API
    import MySQLdb

    try:
        conn = MySQLdb.connect(host=servername, user=username, db=database)
    except MySQLdb.OperationalError(ValueError, NameError) as error:
        log.error(f"Could not connect to Database, {error}")
    else:
        x = conn.cursor()
        try:
            x.execute(sql_command)
        except MySQLdb.Error(ValueError, NameError):
            conn.rollback()
            log.error("MySQL> {0}: {1}".format(sql_command, NameError), ValueError)
        else:
            feedback = x.fetchall()
            conn.commit()
            log.debug("Database query OK: {0}".format(sql_command))
        conn.close()
    return feedback



"""
End-of-night script and functions. Check that everything has been processed,
collect results and merge them if needed.
"""

import logging
import re
import shutil
import subprocess
import sys
from datetime import datetime, timedelta
from pathlib import Path
from typing import Tuple, Iterable, List

from osa import osadb
from osa.configs import options
from osa.configs.config import cfg
from osa.job import are_all_jobs_correctly_finished, save_job_information
from osa.nightsummary.extract import extract_runs, extract_sequences
from osa.nightsummary.nightsummary import run_summary_table
from osa.paths import destination_dir
from osa.raw import is_raw_data_available
from osa.report import start
from osa.utils.cliopts import closercliparsing
from osa.utils.logging import myLogger
from osa.utils.register import register_found_pattern
from osa.utils.utils import (
    night_finished_flag,
    is_day_closed,
    stringify,
    date_to_dir,
    create_lock,
    gettag,
    date_to_iso,
)

__all__ = [
    "is_sequencer_successful",
    "ask_for_closing",
    "post_process",
    "post_process_files",
    "is_finished_check",
    "extract_provenance",
    "merge_dl1_datacheck",
    "set_closed_with_file",
    "merge_files",
    "daily_datacheck",
    "daily_longterm_cmd",
    "observation_finished",
]

log = myLogger(logging.getLogger())


def main():
    """Main function in charge of closing the sequences."""
    closercliparsing()

    if options.verbose:
        log.setLevel(logging.DEBUG)
    else:
        log.setLevel(logging.INFO)

    if options.simulate:
        log.info("Running in simulation mode.")

    # initiating report
    tag = gettag()
    start(tag)

    # starting the algorithm
    if not observation_finished():
        log.warning("Observations not over, it is earlier than 08:00 UTC.")
        sys.exit(0)

    elif is_day_closed():
        log.info(f"Date {date_to_iso(options.date)} already closed for {options.tel_id}")
        sys.exit(0)
    else:
        if options.seqtoclose is not None:
            log.info(f"Closing sequence {options.seqtoclose}")

        sequencer_tuple = [False, []]

        if is_raw_data_available(options.date):
            # proceed normally
            log.debug(f"Checking sequencer_tuple {sequencer_tuple}")
            night_summary_table = run_summary_table(options.date)
            sequencer_tuple = is_finished_check(night_summary_table)

            if not is_sequencer_successful(sequencer_tuple):
                log.info("Sequencer did not complete or finish unsuccessfully")
                ask_for_closing()
        else:
            log.error("Never thought about this possibility, please check the code")
            sys.exit(-1)

        save_job_information()
        post_process(sequencer_tuple)


def is_sequencer_successful(seq_tuple: Tuple[bool, Iterable]):
    """Return a bool assessing whether the sequencer has successfully finished or not."""
    return seq_tuple[0]


def ask_for_closing():
    """
    Ask the user whether sequences should be closed or not.
    A True (Y/y) closes, while False(N/n) answer stops the program.

    Returns
    -------
    object
    """

    if options.noninteractive:
        return

    answer_check = False

    while not answer_check:
        try:
            answer_user = "n"
        except KeyboardInterrupt:
            log.warning("Program exited by user.")
            sys.exit(1)
        except EOFError as error:
            log.exception(f"End of file not expected, {error}")
            sys.exit(2)
        else:
            answer_check = True
            if answer_user in {"n", "N"}:
                # the user does not want to close
                log.info(
                    f"Day {options.date} for {options.tel_id} will "
                    f"remain open unless closing is forced"
                )
                sys.exit(0)
            elif answer_user in {"y", "Y"}:
                continue
            else:
                log.warning("Answer not understood, please type y or n")
                answer_check = False


def post_process(seq_tuple):
    """Set of last instructions."""
    seq_list = seq_tuple[1]

    # Close the sequences
    post_process_files(seq_list)

    # Merge DL1 datacheck files and produce PDFs. It also produces
    # the daily datacheck report using the longterm script, and updates
    # the longterm DL1 datacheck file with the cherenkov_transparency script.
    if cfg.getboolean("lstchain", "merge_dl1_datacheck"):
        list_job_id = merge_dl1_datacheck(seq_list)
        longterm_job_id = daily_datacheck(daily_longterm_cmd(list_job_id))
        cherenkov_transparency(cherenkov_transparency_cmd(longterm_job_id))

    # Extract the provenance info
    extract_provenance(seq_list)

    # Merge DL1b files run-wise
    merge_files(seq_list, data_level="DL1AB")

    merge_muon_files(seq_list)

    # Merge DL2 files run-wise
    if not options.no_dl2:
        merge_files(seq_list, data_level="DL2")

    if options.seqtoclose is None:
        database = cfg.get("database", "path")
        if database:
            osadb.end_processing(date_to_iso(options.date))
        # Creating closing flag files will be deprecated in future versions
        return set_closed_with_file()

    return False


def post_process_files(seq_list: list):
    """
    Identify the different types of files, try to close the sequences
    and copy output files to corresponding data directories.

    Parameters
    ----------
    seq_list: list
        list of sequences
    """

    output_files_set = set(Path(options.directory).rglob("*Run*"))

    DL1AB_RE = re.compile(rf"{options.dl1_prod_id}/dl1.*.(?:h5|hdf5|hdf)")
    MUONS_RE = re.compile(r"muons.*.fits")
    DATACHECK_RE = re.compile(r"datacheck_dl1.*.(?:h5|hdf5|hdf)")
    INTERLEAVED_RE = re.compile(r"interleaved.*.(?:h5|hdf5|hdf)")

    pattern_files = dict(
        [
            ("DL1AB", DL1AB_RE),
            ("MUON", MUONS_RE),
            ("DATACHECK", DATACHECK_RE),
            ("INTERLEAVED", INTERLEAVED_RE),
        ]
    )

    if not options.no_dl2:
        DL2_RE = re.compile(f"{options.dl2_prod_id}/dl2.*.(?:h5|hdf5|hdf)")
        pattern_files["DL2"] = DL2_RE

    for concept, pattern_re in pattern_files.items():
        log.info(f"Post processing {concept} files, {len(output_files_set)} files left")

        dst_path = destination_dir(concept, create_dir=True)

        log.debug(f"Checking if {concept} files need to be moved to {dst_path}")

        for file_path in output_files_set.copy():

            file = str(file_path)
            # If seqtoclose is set, we only want to close that sequence
            if options.seqtoclose is not None and options.seqtoclose not in file:
                continue

            if pattern_found := pattern_re.search(file):
                log.debug(f"Pattern {concept} found, {pattern_found} in {file}")
                registered_file = register_found_pattern(file_path, seq_list, concept, dst_path)
                output_files_set.remove(registered_file)


def set_closed_with_file():
    """Write the analysis report to the closer file."""
    night_finished_file = night_finished_flag()
    is_closed = False
    if not options.simulate:
        # Generate NightFinished lock file
        is_closed = create_lock(night_finished_file)
    else:
        log.debug(f"Simulate the creation of lock file {night_finished_file}")

    return is_closed


def observation_finished(date=datetime.utcnow()) -> bool:
    """
    We consider the observation as finished if it is later
    than 08:00 UTC of the next day set by `options.date`
    """
    next_morning_limit = options.date + timedelta(days=1, hours=8)
    return date > next_morning_limit


def is_finished_check(run_summary):
    """
    Check that all sequences are finished.

    Parameters
    ----------
    run_summary: astropy.Table
        Table containing the run information from a given date.

    Returns
    -------
    seq_finished: bool
        True if all sequences are finished, False otherwise.
    seq_list: list
    """

    sequence_success = False
    if run_summary is not None:
        # building the sequences (the same way as the sequencer)
        run_list = extract_runs(run_summary)
        sequence_list = extract_sequences(options.date, run_list)

        if are_all_jobs_correctly_finished(sequence_list):
            sequence_success = True
        else:
            log.info("Jobs did not correctly/yet finish")

    else:
        # empty file (no sensible data)
        sequence_success = True
        sequence_list = []

    return [sequence_success, sequence_list]


def merge_dl1_datacheck(seq_list) -> List[str]:
    """
    Merge every DL1 datacheck h5 files run-wise and generate the PDF files

    Parameters
    ----------
    seq_list: list of sequence objects
        List of Sequence Objects
    """
    log.debug("Merging dl1 datacheck files and producing PDFs")

    muons_dir = destination_dir("MUON", create_dir=False)
    datacheck_dir = destination_dir("DATACHECK", create_dir=False)

    list_job_id = []

    for sequence in seq_list:
        if sequence.type == "DATA":
            cmd = [
                "sbatch",
                "--parsable",
                "-D",
                options.directory,
                "-o",
                f"log/merge_dl1_datacheck_{sequence.run:05d}_%j.out",
                "-e",
                f"log/merge_dl1_datacheck_{sequence.run:05d}_%j.err",
                "lstchain_check_dl1",
                "--input-file",
                f"{datacheck_dir}/datacheck_dl1_LST-1.Run{sequence.run:05d}.*.h5",
                f"--output-dir={datacheck_dir}",
                f"--muons-dir={muons_dir}",
            ]
            if not options.simulate and not options.test:
                job = subprocess.run(
                    cmd,
                    encoding="utf-8",
                    capture_output=True,
                    text=True,
                    check=True,
                )
                list_job_id.append(job.stdout.strip())
            else:
                log.debug("Simulate launching scripts")

            log.debug(f"Executing {stringify(cmd)}")

    return list_job_id


def extract_provenance(seq_list):
    """
    Extract provenance run wise from the prov.log file
    where it was stored sub-run wise

    Parameters
    ----------
    seq_list: list of sequence objects
        List of Sequence Objects
    """
    log.info("Extract provenance run wise")

    nightdir = date_to_dir(options.date)

    for sequence in seq_list:
        if sequence.type == "DATA":
            drs4_pedestal_run_id = str(sequence.drs4_run)
            pedcal_run_id = str(sequence.pedcal_run)
            cmd = [
                "sbatch",
                "-D",
                options.directory,
                "-o",
                f"log/provenance_{sequence.run:05d}_%j.log",
                "provprocess",
                "-c",
                options.configfile,
                drs4_pedestal_run_id,
                pedcal_run_id,
                f"{sequence.run:05d}",
                nightdir,
                options.prod_id,
            ]
            if options.no_dl2:
                cmd.append("--no-dl2")
                
            if not options.simulate and not options.test and shutil.which("sbatch") is not None:
                subprocess.run(cmd, check=True)
            else:
                log.debug("Simulate launching scripts")


def get_pattern(data_level) -> Tuple[str, str]:
    """Return the subrun wise file pattern for the data level."""
    if data_level == "DL1AB":
        return "dl1_LST-1.Run?????.????.h5", "dl1"
    if data_level == "MUON":
        return "muons_LST-1.Run?????.????.fits", "muon"
    if data_level == "DL2":
        return "dl2_LST-1.Run?????.????.h5", "dl2"

    raise ValueError(f"Unknown data level {data_level}")


def merge_files(sequence_list, data_level="DL2"):
    """Merge DL1b or DL2 h5 files run-wise."""
    log.info(f"Looping over the sequences and merging the {data_level} files")

    data_dir = destination_dir(data_level, create_dir=False)
    pattern, prefix = get_pattern(data_level)

    for sequence in sequence_list:
        if sequence.type == "DATA":
            merged_file = Path(data_dir) / f"{prefix}_LST-1.Run{sequence.run:05d}.h5"

            cmd = [
                "sbatch",
                "-D",
                options.directory,
                "-o",
                f"log/merge_{prefix}_{sequence.run:05d}_%j.log",
                "lstchain_merge_hdf5_files",
                f"--input-dir={data_dir}",
                f"--output-file={merged_file}",
                "--no-image",
                "--no-progress",
                f"--run-number={sequence.run}",
                f"--pattern={pattern}",
            ]

            log.debug(f"Executing {stringify(cmd)}")

            if not options.simulate and not options.test and shutil.which("sbatch") is not None:
                subprocess.run(cmd, check=True)
            else:
                log.debug("Simulate launching scripts")


def merge_muon_files(sequence_list):
    """Merge muon files run-wise."""
    log.info("Looping over the sequences and merging the MUON files")

    data_dir = destination_dir("MUON", create_dir=False)
    pattern, prefix = get_pattern("MUON")

    for sequence in sequence_list:
        merged_file = Path(data_dir) / f"muons_LST-1.Run{sequence.run:05d}.fits"

        cmd = [
            "sbatch",
            "-D",
            options.directory,
            "-o",
            f"log/merge_{prefix}_{sequence.run:05d}_%j.log",
            "lstchain_merge_muon_files",
            f"--input-dir={data_dir}",
            f"--output-file={merged_file}",
            f"--run-number={sequence.run}",
            f"--pattern={pattern}",
        ]

        log.debug(f"Executing {stringify(cmd)}")

        if not options.simulate and not options.test and shutil.which("sbatch") is not None:
            subprocess.run(cmd, check=True)
        else:
            log.debug("Simulate launching scripts")


def daily_longterm_cmd(parent_job_ids: List[str]) -> List[str]:
    """Build the daily longterm command."""
    nightdir = date_to_dir(options.date)
    datacheck_dir = destination_dir("DATACHECK", create_dir=False)
    muons_dir = destination_dir("MUON", create_dir=False)
    longterm_dir = Path(cfg.get("LST1", "LONGTERM_DIR")) / options.prod_id / nightdir
    longterm_output_file = longterm_dir / f"DL1_datacheck_{nightdir}.h5"

    return [
        "sbatch",
        "--parsable",
        "-D",
        options.directory,
        "-o",
        "log/longterm_daily_%j.log",
        f"--dependency=afterok:{','.join(parent_job_ids)}",
        "lstchain_longterm_dl1_check",
        f"--input-dir={datacheck_dir}",
        f"--output-file={longterm_output_file}",
        f"--muons-dir={muons_dir}",
        "--batch",
    ]


def daily_datacheck(cmd: List[str]):
    """Run daily dl1 checks using longterm script."""
    log.info("Daily dl1 checks using longterm script.")
    log.debug(f"Executing {stringify(cmd)}")

    if not options.simulate and not options.test and shutil.which("sbatch") is not None:
        job = subprocess.run(cmd, check=True)
        job_id = job.stdout.strip()
        return job_id
    else:
        log.debug("Simulate launching scripts")


def cherenkov_transparency_cmd(longterm_job_id: str) -> List[str]:
    """Build the cherenkov transparency command."""
    nightdir = date_to_dir(options.date)
    datacheck_dir = destination_dir("DATACHECK", create_dir=False)
    longterm_dir = Path(cfg.get("LST1", "LONGTERM_DIR")) / options.prod_id / nightdir
    longterm_datacheck_file = longterm_dir / f"DL1_datacheck_{nightdir}.h5"

    return [
        "sbatch",
        "-D",
        options.directory,
        "-o",
        "log/cherenkov_transparency_%j.log",
        f"--dependency=afterok:{longterm_job_id}",
        "lstchain_cherenkov_transparency",
        f"--update-datacheck-file={longterm_datacheck_file}",
        f"--input-dir={datacheck_dir}",
    ]


def cherenkov_transparency(cmd: List[str]):
    """Update longterm dl1 check file with cherenkov transparency information."""
    log.info("Update longterm dl1 check file with cherenkov_transparency script.")
    log.debug(f"Executing {stringify(cmd)}")

    if not options.simulate and not options.test and shutil.which("sbatch") is not None:
        subprocess.run(cmd, check=True)
    else:
        log.debug("Simulate launching scripts")


if __name__ == "__main__":
    main()



from freezegun import freeze_time

from osa.configs import options


@freeze_time("2021-12-22 01:02:03")
def test_start_processing_db(database):
    from osa.osadb import start_processing

    options.test = options.simulate = False

    start_processing("2020-01-01")

    database.execute("SELECT * FROM processing WHERE date = ?", ("2020-01-01",))
    assert database.fetchone() == ("LST1", "2020-01-01", "v0.1.0", "2021-12-22 01:02:03", None, 0)
    options.test = options.simulate = True
    assert start_processing("2020-01-01") is None


@freeze_time("2021-12-22 04:05:06")
def test_end_processing_db(database):
    from osa.osadb import end_processing

    options.test = options.simulate = False

    end_processing("2020-01-01")

    database.execute("SELECT * FROM processing WHERE date = ?", ("2020-01-01",))
    assert database.fetchone() == (
        "LST1",
        "2020-01-01",
        "v0.1.0",
        "2021-12-22 01:02:03",
        "2021-12-22 04:05:06",
        1,
    )
    # Test that there is only one entry or row corresponding to the date 2020-01-01
    # even if end_processing is called twice.
    end_processing("2020-01-01")
    database.execute("SELECT * FROM processing WHERE date = ?", ("2020-01-01",))
    query = database.execute("SELECT * FROM processing WHERE date = ?", ("2020-01-01",))
    assert len(query.fetchall()) == 1

    options.test = options.simulate = True

    assert end_processing("2020-01-01") is None


def test_no_osadb():
    from osa.osadb import open_database

    with open_database("/tmp/osa.db") as cursor:
        assert cursor is None



import datetime
import os
import subprocess as sp
from pathlib import Path
from textwrap import dedent

import pytest
import yaml

from osa.configs import options
from osa.scripts.closer import is_sequencer_successful, is_finished_check

ALL_SCRIPTS = [
    "sequencer",
    "closer",
    "copy_datacheck",
    "datasequence",
    "calibration_pipeline",
    "show_run_summary",
    "provprocess",
    "simulate_processing",
    "dl3_stage",
    "theta2_significance",
    "source_coordinates",
    "sequencer_webmaker",
]

options.date = datetime.datetime.fromisoformat("2020-01-17")
options.tel_id = "LST1"
options.prod_id = "v0.1.0"
options.dl1_prod_id = "tailcut84"
options.directory = "test_osa/test_files0/running_analysis/20200117/v0.1.0/"


def remove_provlog():
    log_file = Path("prov.log")
    if log_file.is_file():
        log_file.unlink()


def run_program(*args):
    result = sp.run(args, stdout=sp.PIPE, stderr=sp.STDOUT, encoding="utf-8", check=True)

    if result.returncode != 0:
        new_line = "\n"
        raise ValueError(
            f"Running {args[0]} failed with return code {result.returncode}, output: "
            f"{new_line.join(result.stdout)}"
        )

    return result


@pytest.mark.parametrize("script", ALL_SCRIPTS)
def test_all_help(script):
    """Test for all scripts if at least the help works."""
    run_program(script, "--help")


def test_simulate_processing(
    drs4_time_calibration_files,
    systematic_correction_files,
    run_summary_file,
    r0_data,
    merged_run_summary,
    drive_log
):

    for file in drs4_time_calibration_files:
        assert file.exists()

    for file in systematic_correction_files:
        assert file.exists()

    for r0_file in r0_data:
        assert r0_file.exists()

    assert run_summary_file.exists()
    assert merged_run_summary.exists()
    assert drive_log.exists()

    remove_provlog()
    rc = run_program("simulate_processing", "-p", "--force")
    assert rc.returncode == 0

    prov_dl1_path = Path("./test_osa/test_files0/DL1/20200117/v0.1.0/tailcut84/log")
    prov_dl2_path = Path("./test_osa/test_files0/DL2/20200117/v0.1.0/model2/log")
    prov_file_dl1 = prov_dl1_path / "calibration_to_dl1_01807_prov.log"
    prov_file_dl2 = prov_dl2_path / "calibration_to_dl2_01807_prov.log"
    json_file_dl1 = prov_dl1_path / "calibration_to_dl1_01807_prov.json"
    json_file_dl2 = prov_dl2_path / "calibration_to_dl2_01807_prov.json"
    pdf_file_dl1 = prov_dl1_path / "calibration_to_dl1_01807_prov.pdf"
    pdf_file_dl2 = prov_dl2_path / "calibration_to_dl2_01807_prov.pdf"

    assert prov_file_dl1.exists()
    assert prov_file_dl2.exists()
    assert pdf_file_dl1.exists()
    assert pdf_file_dl2.exists()

    with open(json_file_dl1) as file:
        dl1 = yaml.safe_load(file)
    assert len(dl1["entity"]) == 19
    assert len(dl1["activity"]) == 5
    assert len(dl1["used"]) == 15
    assert len(dl1["wasGeneratedBy"]) == 10

    with open(json_file_dl2) as file:
        dl2 = yaml.safe_load(file)
    assert len(dl2["entity"]) == 25
    assert len(dl2["activity"]) == 6
    assert len(dl2["used"]) == 21
    assert len(dl2["wasGeneratedBy"]) == 12

    rc = run_program("simulate_processing", "-p")
    assert rc.returncode == 0

    remove_provlog()
    rc = run_program("simulate_processing", "-p")
    assert rc.returncode == 0


def test_simulated_sequencer(
    drs4_time_calibration_files,
    systematic_correction_files,
    run_summary_file,
    run_catalog,
    r0_data,
    merged_run_summary,
):
    assert run_summary_file.exists()
    assert run_catalog.exists()

    for r0_file in r0_data:
        assert r0_file.exists()

    for file in drs4_time_calibration_files:
        assert file.exists()

    for file in systematic_correction_files:
        assert file.exists()

    rc = run_program("sequencer", "-d", "2020-01-17", "-s", "-t", "LST1")

    assert rc.returncode == 0
    now = datetime.datetime.now(datetime.timezone.utc).strftime("%Y-%m-%d %H:%M")
    assert rc.stdout == dedent(
        f"""\
        =================================== Starting sequencer.py at {now} UTC for LST, Telescope: LST1, Date: 2020-01-17 ===================================
        Tel   Seq  Parent  Type      Run   Subruns  Source        Action  Tries  JobID  State  CPU_time  Exit  DL1%  MUONS%  DL1AB%  DATACHECK%  DL2%  
        LST1    1  None    PEDCALIB  1809  5        None          None    None   None   None   None      None  None  None    None    None        None  
        LST1    2       1  DATA      1807  11       Crab          None    None   None   None   None      None     0       0       0           0     0  
        LST1    3       1  DATA      1808  9        MadeUpSource  None    None   None   None   None      None     0       0       0           0     0  
        """  # noqa: E501
    )


def test_sequencer(sequence_file_list):
    for sequence_file in sequence_file_list:
        assert sequence_file.exists()


def test_autocloser(running_analysis_dir):
    result = run_program(
        "autocloser",
        "--date",
        "2020-01-17",
        "--test",
        "LST1",
    )
    assert os.path.exists(running_analysis_dir)
    assert result.stdout.split()[-1] == "Exit"


def test_closer(
    r0_data,
    running_analysis_dir,
    test_observed_data,
    run_summary_file,
    drs4_time_calibration_files,
    systematic_correction_files,
    merged_run_summary,
):
    # First assure that the end of night flag is not set and remove it otherwise
    night_finished_flag = Path(
        "./test_osa/test_files0/OSA/Closer/20200117/v0.1.0/NightFinished.txt"
    )
    if night_finished_flag.exists():
        night_finished_flag.unlink()

    for r0_file in r0_data:
        assert r0_file.exists()
    for file in drs4_time_calibration_files:
        assert file.exists()
    for file in systematic_correction_files:
        assert file.exists()
    assert running_analysis_dir.exists()
    assert run_summary_file.exists()
    for obs_file in test_observed_data:
        assert obs_file.exists()
    assert merged_run_summary.exists()

    run_program("closer", "-y", "-v", "-t", "-d", "2020-01-17", "LST1")
    closed_seq_file = running_analysis_dir / "sequence_LST1_01809.closed"

    # Check that files have been moved to their final destinations
    assert os.path.exists(
       "./test_osa/test_files0/DL1/20200117/v0.1.0/muons/muons_LST-1.Run01808.0011.fits"
    )
    assert os.path.exists(
       "./test_osa/test_files0/DL1/20200117/v0.1.0/interleaved/interleaved_LST-1.Run01808.0011.h5"
    )
    assert os.path.exists(
        "./test_osa/test_files0/DL1/20200117/v0.1.0/tailcut84/dl1_LST-1.Run01808.0011.h5"
    )
    assert os.path.exists(
        "./test_osa/test_files0/DL1/20200117/v0.1.0/tailcut84/datacheck/"
        "datacheck_dl1_LST-1.Run01808.0011.h5"
    )
    assert os.path.exists(
        "./test_osa/test_files0/DL2/20200117/v0.1.0/model2/dl2_LST-1.Run01808.0011.h5"
    )
    # Assert that the link to dl1 and muons files have been created
    assert os.path.islink(
        "./test_osa/test_files0/running_analysis/20200117/v0.1.0/muons_LST-1.Run01808.0011.fits"
    )
    assert os.path.islink(
        "./test_osa/test_files0/running_analysis/20200117/v0.1.0/dl1_LST-1.Run01808.0011.h5"
    )

    assert night_finished_flag.exists()
    assert closed_seq_file.exists()


def test_datasequence(running_analysis_dir):
    drs4_file = "drs4_pedestal.Run00001.0000.fits"
    calib_file = "calibration.Run00002.0000.hdf5"
    timecalib_file = "time_calibration.Run00002.0000.hdf5"
    systematic_correction_file = "no_sys_corrected_calibration_scan_fit_20210514.0000.h5"
    drive_file = "DrivePosition_20200117.txt"
    runsummary_file = "RunSummary_20200117.ecsv"
    prod_id = "v0.1.0"
    run_number = "00003.0000"
    options.directory = running_analysis_dir

    output = run_program(
        "datasequence",
        "--date=2020-01-17",
        "--simulate",
        f"--prod-id={prod_id}",
        f"--drs4-pedestal-file={drs4_file}",
        f"--pedcal-file={calib_file}",
        f"--time-calib-file={timecalib_file}",
        f"--systematic-correction-file={systematic_correction_file}",
        f"--drive-file={drive_file}",
        f"--run-summary={runsummary_file}",
        run_number,
        "LST1",
    )
    assert output.returncode == 0


def test_calibration_pipeline(running_analysis_dir):
    options.prod_id = "v0.1.0"
    drs4_run_number = "01804"
    pedcal_run_number = "01805"
    options.directory = running_analysis_dir

    output = run_program(
        "calibration_pipeline",
        "--date=2020-01-17",
        "--simulate",
        f"--prod-id={options.prod_id}",
        f"--drs4-pedestal-run={drs4_run_number}",
        f"--pedcal-run={pedcal_run_number}",
        "LST1",
    )
    assert output.returncode == 0


def test_is_sequencer_successful(run_summary, running_analysis_dir):
    options.directory = running_analysis_dir
    options.test = True
    seq_tuple = is_finished_check(run_summary)
    options.test = False
    assert is_sequencer_successful(seq_tuple) is True


def test_drs4_pedestal_cmd(base_test_dir):
    from osa.scripts.calibration_pipeline import drs4_pedestal_command

    cmd = drs4_pedestal_command(drs4_pedestal_run_id="01804")
    expected_command = [
        "onsite_create_drs4_pedestal_file",
        "--run_number=01804",
        f"--base_dir={base_test_dir}",
        "--no-progress",
    ]
    assert cmd == expected_command


def test_calibration_file_cmd(base_test_dir):
    from osa.scripts.calibration_pipeline import calibration_file_command

    cmd = calibration_file_command(drs4_pedestal_run_id="01804", pedcal_run_id="01809")
    expected_command = [
        "onsite_create_calibration_file",
        "--pedestal_run=01804",
        "--run_number=01809",
        f"--base_dir={base_test_dir}",
    ]
    assert cmd == expected_command


def test_daily_longterm_cmd():
    from osa.scripts.closer import daily_longterm_cmd

    job_ids = ["12345", "54321"]
    cmd = daily_longterm_cmd(parent_job_ids=job_ids)

    expected_cmd = [
        "sbatch",
        "--parsable",
        "-D",
        options.directory,
        "-o",
        "log/longterm_daily_%j.log",
        "--dependency=afterok:12345,54321",
        "lstchain_longterm_dl1_check",
        "--input-dir=test_osa/test_files0/DL1/20200117/v0.1.0/tailcut84/datacheck",
        "--output-file=test_osa/test_files0/OSA/DL1DataCheck_LongTerm/v0.1.0/20200117/DL1_datacheck_20200117.h5",
        "--muons-dir=test_osa/test_files0/DL1/20200117/v0.1.0/muons",
        "--batch",
    ]

    assert cmd == expected_cmd


def test_observation_finished():
    """Check if observation is finished for `options.date=2020-01-17`."""
    from osa.scripts.closer import observation_finished

    date1 = datetime.datetime(2020, 1, 21, 12, 0, 0)
    assert observation_finished(date=date1) is True
    date2 = datetime.datetime(2020, 1, 17, 5, 0, 0)
    assert observation_finished(date=date2) is False


def test_no_runs_found():
    output = sp.run(
        ["sequencer", "-s", "-d", "2015-01-01", "LST1"], text=True, stdout=sp.PIPE, stderr=sp.PIPE
    )
    assert output.returncode == 0
    assert "No runs found for this date. Nothing to do. Exiting." in output.stderr.splitlines()[-1]


@pytest.mark.skip(reason="Currently not working with all combinations")
def test_sequencer_webmaker(
    run_summary,
    merged_run_summary,
    drs4_time_calibration_files,
    systematic_correction_files,
    base_test_dir,
):
    # Check if night finished flag is set
    night_finished = base_test_dir / "OSA/Closer/20200117/v0.1.0/NightFinished.txt"

    if night_finished.exists():
        output = sp.run(
            ["sequencer_webmaker", "--test", "-d", "2020-01-17"],
            text=True,
            stdout=sp.PIPE,
            stderr=sp.PIPE,
        )
        assert output.returncode != 0
        assert output.stderr.splitlines()[-1] == "Date 2020-01-17 is already closed for LST1"
        night_finished.unlink()

    output = sp.run(["sequencer_webmaker", "--test", "-d", "2020-01-17"])
    assert output.returncode == 0
    directory = base_test_dir / "OSA" / "SequencerWeb"
    directory.mkdir(parents=True, exist_ok=True)
    expected_file = directory / "osa_status_20200117.html"
    assert expected_file.exists()

    output = sp.run(["sequencer_webmaker", "--test"])
    assert output.returncode != 0

    # Running without test option will make the script fail
    output = sp.run(["sequencer_webmaker", "-d", "2020-01-17"])
    assert output.returncode != 0



"""Handle the list of closed and vetoed sequences."""

import logging
import os
from pathlib import Path

from osa.configs import options
from osa.utils.logging import myLogger

__all__ = [
    "failed_history",
    "get_veto_list",
    "get_closed_list",
    "set_veto_action",
    "set_closed_action",
    "update_vetoes",
    "set_closed_sequence",
]

log = myLogger(logging.getLogger(__name__))


def get_veto_list(sequence_list):
    """Get a list of vetoed sequences."""
    update_vetoes(sequence_list)
    analysis_dir = Path(options.directory)
    veto_ls = analysis_dir.glob("*.veto")
    veto_list = []
    for file in veto_ls:
        # Extract the job name
        name = file.stem.strip("sequence_")
        veto_list.append(name)
        set_veto_action(name, sequence_list)
    return veto_list


def set_veto_action(name, sequence_list):
    """Set the action for a given sequence to veto."""
    for sequence in sequence_list:
        if sequence.jobname == name:
            sequence.action = "Veto"
            log.debug(f"Attributes of sequence {sequence.seq} updated")


def update_vetoes(sequence_list):
    """Create a .veto file for a given sequence if reached maximum number of trials."""
    for sequence in sequence_list:
        if (
            not os.path.exists(sequence.veto)
            and os.path.exists(sequence.history)
            and failed_history(sequence.history)
        ):
            Path(sequence.veto).touch()
            log.debug(f"Created veto file {sequence.veto}")


def failed_history(history_file: Path) -> bool:
    """
    Check if a processing step has failed twice in a given history file.

    Return True if the last line of the history file contains a non-zero exit
    status and is repeated twice, meaning that a given step has failed twice.
    """
    history_lines = history_file.read_text().splitlines()

    # Check if history file has at least two trials
    if len(history_lines) < 2:
        return False

    # Check if the last line of the history file is repeated twice
    # and the exit status is non-zero
    return history_lines[-1] == history_lines[-2] and history_lines[-1].split()[-1] != "0"


def set_closed_sequence(sequence):
    """Creates a .closed lock file for a given sequence."""
    sequence.closed = Path(options.directory) / f"sequence_{sequence.jobname}.closed"
    sequence.closed.touch()


def get_closed_list(sequence_list) -> list:
    """Get the list of closed sequences."""
    analysis_dir = Path(options.directory)
    closed_ls = analysis_dir.glob("*.closed")
    closed_list = []
    for file in closed_ls:
        # Extract the job name: LST1_XXXXX
        name = file.stem.strip("sequence_")
        closed_list.append(name)
        set_closed_action(name, sequence_list)
    return closed_list


def set_closed_action(name: str, sequence_list):
    """Set the action of a closed sequence object."""
    for sequence in sequence_list:
        if sequence.jobname == name:
            sequence.action = "Closed"
            log.debug(f"Attributes of sequence {sequence.seq} updated")



import subprocess as sp


def test_dl3_stage():
    output = sp.run(
        ["dl3_stage", "-d", "2020-01-17", "-s", "LST1"], text=True, stdout=sp.PIPE, stderr=sp.PIPE
    )
    assert output.returncode == 0
    assert "Creating observation index for each source." in output.stderr.splitlines()[-1]



import numpy as np
from astropy.coordinates import SkyCoord
from astropy.table import Table

from osa.scripts.tests.test_osa_scripts import run_program


def test_source_coordinates(run_catalog):
    assert run_catalog.exists()

    output = run_program(
        "source_coordinates",
        "--source",
        "MadeUpSource",
        "--ra",
        "31.22",
        "--dec",
        "52.95",
        str(run_catalog),
    )

    assert output.returncode == 0
    table = Table.read(run_catalog)

    # Check new coordinates
    crab_coords = SkyCoord.from_name("Crab")
    assert np.isclose(table[table["source_name"] == "Crab"]["source_ra"], crab_coords.ra.deg)
    assert np.isclose(table[table["source_name"] == "Crab"]["source_dec"], crab_coords.dec.deg)
    assert np.isclose(table[table["source_name"] == "MadeUpSource"]["source_ra"], 31.22)
    assert np.isclose(table[table["source_name"] == "MadeUpSource"]["source_dec"], 52.95)



"""Query the TCU database source name and astronomical coordinates."""
import logging
from datetime import datetime
from typing import Tuple

from pymongo import MongoClient
from pymongo.errors import ConnectionFailure

from osa.configs.config import cfg
from osa.utils.logging import myLogger

__all__ = ["query", "db_available", "get_run_info_from_TCU"]


log = myLogger(logging.getLogger(__name__))


CACO_DB = cfg.get("database", "caco_db")
TCU_DB = cfg.get("database", "tcu_db")


def db_available():
    """Check the connection to the TCU database."""
    caco_client = MongoClient(CACO_DB, serverSelectionTimeoutMS=3000)
    tcu_client = MongoClient(TCU_DB, serverSelectionTimeoutMS=3000)
    try:
        caco_client.server_info()
        tcu_client.server_info()
    except ConnectionFailure:
        log.warning("TCU or CaCo database not available. No source info will be added.")
        return False
    else:
        log.debug("TCU and CaCo database are available. Source info will be added.")
        return True


def query(obs_id: int, property_name: str):
    """
    Query the source name and coordinates from TCU database.

    Parameters
    ----------
    obs_id : int
        Run number
    property_name : str
        Properties from drive information e.g. `DriveControl_SourceName`,
        `DriveControl_RA_Target`, `DriveControl_Dec_Target`

    Returns
    -------
    query_result : str or None
        Query result from database. It can be either the source name or its coordinates.

    Raises
    ------
    ConnectionFailure
    """

    # Avoid problems with numpy int64 encoding in MongoDB
    if not isinstance(obs_id, int):
        obs_id = int(obs_id)

    caco_client = MongoClient(CACO_DB)
    tcu_client = MongoClient(TCU_DB)

    with caco_client, tcu_client:
        run_info = caco_client["CACO"]["RUN_INFORMATION"]
        run = run_info.find_one({"run_number": obs_id})

        try:
            start = datetime.fromisoformat(run["start_time"].replace("Z", ""))
            end = datetime.fromisoformat(run["stop_time"].replace("Z", ""))
        except TypeError:
            return None

        bridges_monitoring = tcu_client["bridgesmonitoring"]
        property_collection = bridges_monitoring["properties"]
        chunk_collection = bridges_monitoring["chunks"]
        descriptors = property_collection.find(
            {"property_name": property_name},
        )

        entries = {"name": property_name, "time": [], "value": []}

        for descriptor in descriptors:
            query_property = {"pid": descriptor["_id"]}

            if start is not None:
                query_property["begin"] = {"$gte": start}

            if end is not None:
                query_property["end"] = {"$lte": end}

            chunks = chunk_collection.find(query_property)

            for chunk in chunks:
                for value in chunk["values"]:
                    entries["time"].append(value["t"])
                    entries["value"].append(value["val"])

                    source_name = entries["value"][0]
                    return source_name if source_name != "" else None


def get_run_info_from_TCU(run_id: int, tcu_server: str) -> Tuple:
    """
    Get type of run, start, end timestamps (in iso format)
    and elapsed time (in minutes) for a given run ID.

    Parameters
    ----------
    run_id: int
    tcu_server: str
        Host of the TCU database

    Returns
    -------
    run_id : int
    run_type : str
    tstart
    tstop
    elapsed : float
        Elapsed time in minutes

    Notes
    -----
    TCU Monitoring database can be accessed at 'lst101-int' server
    from lstanalyzer@tcs06 and cp's
    """

    client = MongoClient(tcu_server)
    collection = client["lst1_obs_summary"]["camera"]
    summary = collection.find_one({"run_number": run_id})

    if summary is not None:
        run_type = summary["kind"]
        tstart = datetime.fromtimestamp(summary["tstart"]).isoformat(sep=" ", timespec="seconds")
        tstop = datetime.fromtimestamp(summary["tstop"]).isoformat(sep=" ", timespec="seconds")
        elapsed = (summary["tstop"] - summary["tstart"]) / 60  # minutes

        return run_id, run_type, tstart, tstop, elapsed

    return run_id, None, None, None, 0.0



import logging
from datetime import datetime, timezone
from fnmatch import fnmatchcase
from glob import glob
from os.path import basename, getsize, join
from pathlib import Path

from osa.configs import config, options
from osa.configs.config import cfg
from osa.raw import get_raw_dir
from osa.utils.iofile import append_to_file
from osa.utils.logging import myLogger
from osa.utils.utils import date_to_iso

log = myLogger(logging.getLogger(__name__))

__all__ = ["history", "start", "finished_assignments"]


def start(parent_tag: str):
    """Print out the header of the script (e.g. sequencer, closer)."""
    now = datetime.now(timezone.utc)
    simple_parent_tag = parent_tag.rsplit("(")[0]
    header(
        f"Starting {simple_parent_tag} at {now.strftime('%Y-%m-%d %H:%M')} "
        f"UTC for LST, Telescope: {options.tel_id}, "
        f"Date: {options.date.strftime('%Y-%m-%d')}"
    )


def header(message):
    """Print out a header of a given length."""
    framesize = int(config.cfg.get("OUTPUT", "REPORTWIDTH"))
    if len(message) < framesize - 2:
        prettyframe = int((framesize - 2 - len(message)) / 2) * "="
    else:
        prettyframe = ""
    log.info(f"{prettyframe} {message} {prettyframe}")


def finished_assignments(sequence_list):
    """
    Report that the files have been produced and the night closed.

    Parameters
    ----------
    sequence_list: Iterable

    Returns
    -------
    dictionary: dict
        Dictionary with the sequence names as keys and the corresponding
        output files and sizes as values.
    """
    concept_set = []
    analysis_dir = options.directory
    disk_space_GB = 0
    rawnum = 0
    if options.tel_id == "LST1":
        concept_set = [
            "PEDESTAL",
            "CALIB",
            "TIMECALIB",
            "DL1",
            "DL1AB",
            "MUON",
            "DATACHECK",
            "DL2",
        ]
        rawdir = get_raw_dir(options.date)
        if sequence_list is not None:
            for seq in sequence_list:
                rawnum += seq.subruns
        data_files = glob(
            join(
                rawdir,
                f'*{cfg.get("PATTERN", "R0PREFIX")}*{cfg.get("PATTERN", "R0SUFFIX")}*',
            )
        )
        disk_space = sum(getsize(d) for d in data_files)
        disk_space_GB_f = float(disk_space) / (1000 * 1000 * 1000)
        disk_space_GB = int(round(disk_space_GB_f, 0))

    ana_files = glob(join(analysis_dir, "*" + cfg.get("PATTERN", "R0SUFFIX")))
    file_no = {}
    ana_set = set(ana_files)

    for concept in concept_set:
        pattern = f"{cfg.get('PATTERN', f'{concept}PREFIX')}*"
        log.debug(f"Trying with {concept} and searching {pattern}")
        file_no[concept] = 0
        delete_set = set()
        for a in ana_set:
            ana_file = basename(a)
            if pattern_found := fnmatchcase(ana_file, pattern):
                log.debug(f"Was pattern {pattern} found in {ana_file}?: {pattern_found}")
                file_no[concept] += 1
                delete_set.add(a)
        ana_set -= delete_set

    now_string = f"{datetime.now(timezone.utc).strftime('%Y-%m-%d %H:%M')}"

    dictionary = {
        "NIGHT": date_to_iso(options.date),
        "TELESCOPE": options.tel_id,
        "IS_CLOSED": 1,
        "SEQUENCES": len(sequence_list),
        "FILES_RAW": rawnum,
        "RAW_GB": disk_space_GB,
        "END": now_string,
    }

    for concept in concept_set:
        dictionary[f"FILES_{concept}"] = file_no[concept]

    return dictionary


def history(
    run: str,
    prod_id: str,
    stage: str,
    return_code: int,
    history_file: Path,
    input_file=None,
    config_file=None,
) -> None:
    """
    Appends a history line to the history file. A history line
    reports the outcome of the execution of a lstchain executable.

    Parameters
    ----------
    run : str
        Run/sequence analyzed.
    prod_id : str
        Prod ID of the run analyzed.
    stage : str
        Stage of the analysis pipeline.
    return_code : int
        Return code of the lstchain executable.
    history_file : pathlib.Path
        The history file that keeps track of the analysis steps.
    input_file : str, optional
        If needed, input file used for the lstchain executable
    config_file : str, optional
        Input card used for the lstchain executable.
    """
    date_string = datetime.utcnow().isoformat(sep=" ", timespec="minutes")
    string_to_write = (
        f"{run} {stage} {prod_id} {date_string} " f"{input_file} {config_file} {return_code}\n"
    )
    append_to_file(history_file, string_to_write)



"""Definition of classes containing run and sequence information"""

from dataclasses import dataclass


@dataclass
class RunObj:
    run_str: str = None
    run: int = None
    type: str = None
    subruns: int = None
    source_name: str = None
    source_ra: float = None
    source_dec: float = None
    telescope: str = "LST1"
    night: str = None


class Sequence(RunObj):
    def __init__(self):
        super(Sequence, self).__init__()
        self.seq = None
        self.pedcal_run = None
        self.drs4_run = None
        self.parent = None
        self.previousrun = None
        self.script = None
        self.veto = None
        self.closed = None
        self.history = None
        self.queue = None
        self.jobname = None
        self.jobid = None
        self.action = None
        self.tries = None
        self.state = None
        self.jobhost = None
        self.cputime = None
        self.walltime = None
        self.exit = None

    def associate(self, r):
        for a in r.__dict__.keys():
            self.__dict__.update({a: r.__dict__[a]})


class SequenceCalibration(Sequence):
    def __init__(self, r):
        super(SequenceCalibration, self).__init__()
        super(SequenceCalibration, self).associate(r)
        self.calibstatus = None
        self.seq = 1
        self.parent = None
        self.drs4_run = None


class SequenceData(Sequence):
    def __init__(self, r):
        super(SequenceData, self).__init__()
        super(SequenceData, self).associate(r)
        self.parent = 1
        self.time_calibration_run = None
        self.drive = None
        self.drs4_file = None
        self.calibration_file = None
        self.time_calibration_file = None
        self.systematic_correction_file = None
        self.dl1status = None
        self.dl1abstatus = None
        self.muonstatus = None
        self.datacheckstatus = None
        self.dl2status = None
        self.dl3status = None



"""Utility functions for OSA pipeline provenance."""


from importlib.resources import files
from pathlib import Path

from osa.configs import options
from osa.configs.config import cfg
from osa.utils.utils import date_to_dir, get_lstchain_version

__all__ = ["parse_variables", "get_log_config"]

REDUCTION_TASKS = ["r0_to_dl1", "dl1ab", "dl1_datacheck", "dl1_to_dl2"]


def parse_variables(class_instance):
    """Parse variables needed in model"""

    # calibration_pipeline.py
    # -c sequencer.cfg
    # -d 2020_02_18
    # --drs4-pedestal-run 01804
    # --pedcal-run 01805
    # LST1

    # datasequence.py
    # -c sequencer.cfg
    # -d 2020_02_18
    # --prod-id v0.4
    # --pedcal-file .../20200218/v0.4.3/calibration_filters_52.Run02006.0000.h5
    # --drs4-pedestal-file .../20200218/v0.4.3/drs4_pedestal.Run02005.0000.h5
    # --time-calib-file .../20191124/pro/time_calibration.Run01625.0000.h5
    # --systematic_correction_file .../20200725/pro/ffactor_systematics_20200725.h5
    # --drive-file .../DrivePositioning/DrivePosition_20200218.txt
    # --run-summary .../monitoring/RunSummary/RunSummary_20200101.ecsv
    # 02006.0000
    # LST1

    flat_date = date_to_dir(options.date)
    configfile_dl1b = cfg.get("lstchain", "dl1b_config")
    configfile_dl2 = cfg.get("lstchain", "dl2_config")
    raw_dir = Path(cfg.get("LST1", "R0_DIR"))
    rf_models_directory = Path(cfg.get("lstchain", "RF_MODELS"))
    dl1_dir = Path(cfg.get("LST1", "DL1_DIR"))
    dl2_dir = Path(cfg.get("LST1", "DL2_DIR"))
    calib_dir = Path(cfg.get("LST1", "CALIB_DIR"))
    pedestal_dir = Path(cfg.get("LST1", "PEDESTAL_DIR"))

    class_instance.SoftwareVersion = get_lstchain_version()
    class_instance.ProcessingConfigFile = str(options.configfile)
    class_instance.ObservationDate = flat_date
    if class_instance.__name__ in REDUCTION_TASKS:
        muon_dir = dl1_dir / flat_date / options.prod_id / "muons"
        outdir_dl1 = dl1_dir / flat_date / options.prod_id / options.dl1_prod_id
        outdir_dl2 = dl2_dir / flat_date / options.prod_id / options.dl2_prod_id

    if class_instance.__name__ in ["drs4_pedestal", "calibrate_charge"]:
        # drs4_pedestal_run_id  [0] 1804
        # pedcal_run_id         [1] 1805
        # history_file          [2] .../20210913/v0.7.5/sequence_LST1_01805.0000.history
        class_instance.PedestalRun = f"{class_instance.args[0]:05d}"
        class_instance.CalibrationRun = f"{class_instance.args[1]:05d}"

        version = get_lstchain_version()

        # according to code in onsite scripts in lstchain
        class_instance.RawObservationFilePedestal = str((
            raw_dir / flat_date / f"LST-1.1.Run{class_instance.args[0]:05d}.fits.fz"
        ).resolve())
        class_instance.RawObservationFileCalibration = str((
            raw_dir / flat_date / f"LST-1.1.Run{class_instance.args[1]:05d}.fits.fz"
        ).resolve())
        class_instance.PedestalCheckPlot = str((
            pedestal_dir
            / flat_date
            / version
            / f"log/drs4_pedestal.Run{class_instance.args[0]:05d}.0000.pdf"
        ).resolve())
        class_instance.CalibrationCheckPlot = str((
            calib_dir
            / flat_date
            / version
            / f"log/calibration_filters_52.Run{class_instance.args[1]:05d}.0000.pdf"
        ).resolve())

        # according to code in sequence_calibration_files from paths.py
        class_instance.PedestalFile = str((
            pedestal_dir
            / flat_date
            / version
            / f"drs4_pedestal.Run{class_instance.args[0]:05d}.0000.h5"
        ).resolve())
        class_instance.CoefficientsCalibrationFile = str((
            calib_dir
            / flat_date
            / version
            / f"calibration_filters_52.Run{class_instance.args[1]:05d}.0000.h5"
        ).resolve())

    if class_instance.__name__ == "r0_to_dl1":
        # calibration_file   [0] .../20200218/v0.4.3/calibration_filters_52.Run02006.0000.h5
        # drs4_pedestal_file [1] .../20200218/v0.4.3/drs4_pedestal.Run02005.0000.h5
        # time_calib_file    [2] .../20191124/v0.4.3/time_calibration.Run01625.0000.h5
        # systematic_corr    [3] .../20200101/v0.4.3/no_sys_corrected_calib_20210514.0000.h5
        # drive_file         [4] .../DrivePositioning/DrivePosition_20200218.txt
        # run_summary_file   [5] .../RunSummary/RunSummary_20200101.ecsv
        # pedestal_ids_file  [6] .../path/to/interleaved/pedestal/events.h5
        # run_str            [7] 02006.0000

        run_subrun = class_instance.args[7]
        run = run_subrun.split(".")[0]
        class_instance.ObservationRun = run

        calibration_file = Path(class_instance.args[0]).resolve()
        pedestal_file = Path(class_instance.args[1]).resolve()
        timecalibration_file = Path(class_instance.args[2]).resolve()
        systematic_correction_file = Path(class_instance.args[3]).resolve()
        class_instance.R0SubrunDataset = str((
            raw_dir / flat_date / f"LST-1.1.Run{run_subrun}.fits.fz"
        ).resolve())
        class_instance.CoefficientsCalibrationFile = str(calibration_file)
        class_instance.PedestalFile = str(pedestal_file)
        class_instance.TimeCalibrationFile = str(timecalibration_file)
        class_instance.SystematicCorrectionFile = str(systematic_correction_file)
        class_instance.PointingFile = str(Path(class_instance.args[4]).resolve())
        class_instance.RunSummaryFile = str(Path(class_instance.args[5]).resolve())
        class_instance.DL1SubrunDataset = str(
            (outdir_dl1 / f"dl1_LST-1.Run{run_subrun}.h5").resolve()
        )
        class_instance.MuonsSubrunDataset = str((
            muon_dir / f"muons_LST-1.Run{run_subrun}.fits"
        ).resolve())
        class_instance.InterleavedPedestalEventsFile = None
        if class_instance.args[6] is not None:
            class_instance.InterleavedPedestalEventsFile = str(Path(class_instance.args[6]))

    if class_instance.__name__ == "dl1ab":
        # run_str       [0] 02006.0000

        class_instance.Analysisconfigfile_dl1 = str(Path(configfile_dl1b))
        class_instance.ObservationRun = class_instance.args[0].split(".")[0]
        class_instance.StoreImage = cfg.getboolean("lstchain", "store_image_dl1ab")
        class_instance.DL1SubrunDataset = str((
            outdir_dl1 / f"dl1_LST-1.Run{class_instance.args[0]}.h5"
        ).resolve())

    if class_instance.__name__ == "dl1_datacheck":
        # run_str       [0] 02006.0000
        run_subrun = class_instance.args[0]
        run = run_subrun.split(".")[0]

        class_instance.ObservationRun = run
        class_instance.DL1SubrunDataset = str(
            (outdir_dl1 / f"dl1_LST-1.Run{run_subrun}.h5").resolve()
        )
        class_instance.MuonsSubrunDataset = str((
            muon_dir / f"muons_LST-1.Run{run_subrun}.fits"
        ).resolve())
        class_instance.DL1CheckSubrunDataset = str((
            outdir_dl1 / f"datacheck_dl1_LST-1.Run{run_subrun}.h5"
        ).resolve())
        class_instance.DL1CheckHDF5File = str((
            outdir_dl1 / f"datacheck_dl1_LST-1.Run{run}.h5"
        ).resolve())
        class_instance.DL1CheckPDFFile = str((
            outdir_dl1 / f"datacheck_dl1_LST-1.Run{run}.pdf"
        ).resolve())

    if class_instance.__name__ == "dl1_to_dl2":
        # run_str       [0] 02006.0000
        run_subrun = class_instance.args[0]
        run = run_subrun.split(".")[0]

        class_instance.Analysisconfigfile_dl2 = configfile_dl2
        class_instance.ObservationRun = run
        class_instance.RFModelEnergyFile = str((rf_models_directory / "reg_energy.sav").resolve())
        class_instance.RFModelDispNormFile = str(
            (rf_models_directory / "reg_disp_norm.sav").resolve()
        )
        class_instance.RFModelDispSignFile = str(
            (rf_models_directory / "reg_disp_sign.sav").resolve()
        )
        class_instance.RFModelGammanessFile = str((rf_models_directory / "cls_gh.sav").resolve())
        class_instance.DL1SubrunDataset = str(
            (outdir_dl1 / f"dl1_LST-1.Run{run_subrun}.h5").resolve()
        )
        class_instance.DL2SubrunDataset = str(
            (outdir_dl2 / f"dl2_LST-1.Run{run_subrun}.h5").resolve()
        )
        class_instance.DL2MergedFile = str((outdir_dl2 / f"dl2_LST-1.Run{run}.h5").resolve())

    if class_instance.__name__ in REDUCTION_TASKS:
        class_instance.session_name = class_instance.ObservationRun

    return class_instance


def get_log_config():
    """Get logging configuration from provenance logger config file."""
    std_logger_file = files("osa.provenance") / "config/logger.yaml"
    return std_logger_file.read_text()
